/*
	 Copyright (C) 2003 Paul Brossier

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License as published by
	 the Free Software Foundation; either version 2 of the License, or
	 (at your option) any later version.

	 This program is distributed in the hope that it will be useful,
	 but WITHOUT ANY WARRANTY; without even the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	 GNU General Public License for more details.

	 You should have received a copy of the GNU General Public License
	 along with this program; if not, write to the Free Software
	 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	 
*/

#include "aubio_priv.h"
#include "fft.h"

#include <fftw3.h>
#define FFTW_TYPE fftwf_complex


struct _aubio_fft_t {
	int fft_size;
	int channels;
	float 				*in, *out;
	fft_data_t 		*specdata;
	fftwf_plan 		pfw, pbw;
};

// static functions called by creat/delete
static aubio_fft_t * spectrum_alloc(int size);
static void spectrum_free(aubio_fft_t *s);

/**
 * static objects to calculate fft. 
 * these works as scratch pads for all fft executions
 */

static aubio_fft_t * spectrum_alloc(int size) {
	aubio_fft_t * s = AUBIO_NEW(aubio_fft_t);
	// allocate memory
	s->in				= AUBIO_ARRAY(float,size);
	s->specdata	= (fft_data_t*)fftwf_malloc(sizeof(fft_data_t)*size);
	s->out			= AUBIO_ARRAY(float,size);
	return s;
}

static void spectrum_free(aubio_fft_t * s) { 
	// destroy data
	fftwf_destroy_plan(s->pfw);
	fftwf_destroy_plan(s->pbw);
	if (s->specdata) 	fftwf_free(s->specdata);
	if (s->out)				AUBIO_FREE(s->out);
	if (s->in )				AUBIO_FREE(s->in );
}

aubio_fft_t * new_aubio_fft(int size) {
	aubio_fft_t * s =(aubio_fft_t *)spectrum_alloc(size);
	// create plans
	s->pfw = fftwf_plan_dft_r2c_1d(size, s->in,  s->specdata, FFTW_ESTIMATE);
	s->pbw = fftwf_plan_dft_c2r_1d(size, s->specdata, s->out, FFTW_ESTIMATE);
	return s;
}

void del_aubio_fft(aubio_fft_t * s) {
	spectrum_free(s);
}

void aubio_fft_do(const aubio_fft_t * s, 
		const float * data, fft_data_t * spectrum, 
		const int size) {
	unsigned int i;
	for (i=0;i<size;i++) s->in[i] = data[i];
	fftwf_execute(s->pfw);
	for (i=0;i<size;i++) spectrum[i] = s->specdata[i];
}

void aubio_fft_rdo(const aubio_fft_t * s, 
		const fft_data_t * spectrum, 
		float * data, 
		const int size) {
	unsigned int i;
	const float renorm = 1./(float)size;
	for (i=0;i<size;i++) s->specdata[i] = spectrum[i];
	fftwf_execute(s->pbw);
	for (i=0;i<size;i++) data[i] = s->out[i]*renorm;
}

