/*
	 Copyright (C) 2003 Paul Brossier

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License as published by
	 the Free Software Foundation; either version 2 of the License, or
	 (at your option) any later version.

	 This program is distributed in the hope that it will be useful,
	 but WITHOUT ANY WARRANTY; without even the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	 GNU General Public License for more details.

	 You should have received a copy of the GNU General Public License
	 along with this program; if not, write to the Free Software
	 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* see in mathutils.h for doc */

#include "aubio_priv.h"
#include "sample.h"
#include "mathutils.h"

/** sort peak comparison function */
static int sort_peak_comp(const void *x, const void *y);
static int sort_array_comp(const void *x, const void *y);

void window(float *w, int size, window_type_t wintype) {
	int i;
	switch(wintype) {
		case rectangle:
			for (i=0;i<size;i++)
				w[i] = 0.5; 
			break;
		case hamming:
			for (i=0;i<size;i++)
				w[i] = 0.54 - 0.46 * COS(TWO_PI * i / (size));
			break;
		case hanning:
			for (i=0;i<size;i++)
				w[i] = 0.5 - (0.5 * COS(TWO_PI * i / (size)));
			break;
		case hanningz:
			for (i=0;i<size;i++)
				w[i] = 0.5 * (1.0 - COS(TWO_PI * i / (size)));
			break;
		case blackman:
			for (i=0;i<size;i++)
				w[i] = 0.42
					- 0.50 * COS(    TWO_PI*i/(size-1.0))
					+	0.08 * COS(2.0*TWO_PI*i/(size-1.0));
			break;
		case blackman_harris:
			for (i=0;i<size;i++)
				w[i] = 0.35875 
					- 0.48829 * COS(    TWO_PI*i/(size-1.0))
					+ 0.14128 * COS(2.0*TWO_PI*i/(size-1.0))
					- 0.01168 * COS(3.0*TWO_PI*i/(size-1.0));
			break;
		case gaussian:
			for (i=0;i<size;i++)
				w[i] = EXP(- 1.0 / SQR(size) * SQR(2.0*i-size));
			break;
		case welch:
			for (i=0;i<size;i++)
				w[i] = 1.0 - SQR((2*i-size)/(size+1.0));
			break;
		case parzen:
			for (i=0;i<size;i++)
				w[i] = 1.0 - fabsf((2*i-size)/(size+1.0));
			break;
		default:
			break;
	}
}

void getnorm(float * norm, fft_data_t * spectrum, int size) {
	int i;
	int tmpsize = size / 2 ;
	for (i=0;i<tmpsize;i++) {
		norm[i]      = ABSC(spectrum[i]);
		//norm[size-i] = norm[i];
	}
	//norm[tmpsize]=0.;
}

void getphase(float * phas, fft_data_t * spectrum, int size) {
	int i;
	int tmpsize = size / 2 ;
	for (i=0;i<tmpsize;i++) {
		phas[i] 			= +ARGC(spectrum[i]);
		//phas[size-i]  = -phas[i];
	}
	//phas[tmpsize]=PI;
}

float unwrap2pi(float phase) {
	// mod(phase+pi,-2pi)+pi
	return phase + TWO_PI * (1. + floorf(-(phase+PI)/TWO_PI));
}

/* cubic interpolation  */
void interpolfreq(
		const float * r, const float * phi, 
		const int N, 
		const spectral_peak_t * ft, const int nbins,
		const int zp, 
		float * iftloc, float * iftphase, float * iftval)
{
	unsigned int i;
	for (i=0;i<nbins;i++) {
		unsigned int tmpbin;
		float leftftval, rightftval, tmpftval, leftftphase, rightftphase, 
		intpfactor, diffphase;
		/* calculate interpolated peak position in bins (iftloc) */
		tmpbin			= (ft[i].loc-1)*((ft[i].loc-1)>0)+((ft[i].loc-1)<=0)*1;
		leftftval	  = r[tmpbin];
		tmpbin			=	(ft[i].loc+1)*((ft[i].loc+1)<N/2)+((ft[i].loc+1)>=N/2)*(N/2);
		rightftval	= r[tmpbin];
		leftftval 	= LIN2DB(leftftval);
		rightftval	= LIN2DB(rightftval);
		tmpftval 		= LIN2DB(ft[i].val);
		iftloc[i]		= ft[i].loc + 0.5*(leftftval-rightftval) / 
			(leftftval - 2.0*tmpftval + rightftval);
		iftloc[i] 	= (iftloc[i]>=1)*iftloc[i] + (iftloc[i]<1)*1;
		iftloc[i] 	= (iftloc[i]>N/2+1)*(zp/2+1) + (iftloc[i]<=N/2+1)*iftloc[i];
		/* calculate interpolated phase (iphase) */
		leftftphase = phi[(int)floor(iftloc[i])];
		rightftphase= phi[(int)floor(iftloc[i])+1];
		intpfactor 	= iftloc[i]-ft[i].loc;
		intpfactor 	= (intpfactor>0)*intpfactor 
			+ (intpfactor<0)*(1+intpfactor);
		diffphase 	= unwrap2pi(rightftphase-leftftphase);
		iftphase[i] = leftftphase+intpfactor*diffphase;
		/* calculate interpolate amplitude (iftval) */
		iftval[i] 	= tmpftval-0.25*(leftftval-rightftval)*(iftloc[i]-ft[i].loc);
		iftval[i]		= DB2LIN(iftval[i]);
	}
}

void peak_pick(
		const float * r, 
		int nbins, const int N, const int threshold,
		spectral_peak_t * ft
		)
{
	unsigned int i, curbin = 0;
	for (i=1;i<N/2-1;i++) {
		if ((r[i]>r[i+1]) &&
				(r[i]>r[i-1]) &&
				(r[i]>threshold)) {
			ft[curbin].loc = i;
			ft[curbin].val = r[i];
		}	else {
			ft[curbin].loc = i;
			ft[curbin].val = 0;
		}
		curbin++;
		if (curbin>nbins) break;
	}	
}


float vec_median(fvec_t *t) {
	int half = (int)floorf((float)t->length/2.);
	sort_array(t);
	//return (t->data[0][half] + t->data[0][half+1]) / 2.;
	return t->data[0][half];
	
}

void sort_array(fvec_t * t) {
	unsigned int i;
	for (i=0;i<t->channels;i++) 
		qsort(t->data[i], t->length, sizeof(float), sort_array_comp);
}

static int sort_array_comp(const void *x, const void *y) {
	if (x < y)  return 1;
	else  			return -1;
}


void sort_peak(spectral_peak_t * ft, int nbins) {
	qsort(ft, nbins, sizeof(spectral_peak_t), sort_peak_comp);
}

static int sort_peak_comp(const void *x, const void *y) {
	spectral_peak_t * ft1 = (spectral_peak_t *) x;
	spectral_peak_t * ft2 = (spectral_peak_t *) y;
	if (ft1->val < ft2->val) 
		return 1;
	else  
		return -1;
}


float vec_mean(fvec_t *s) 
{
	unsigned int i,j;
	float tmp = 0.0f;
	for (i=0; i < s->channels; i++)
		for (j=0; j < s->length; j++)
			tmp += s->data[i][j];
	return tmp/(float)(s->length);
}


float vec_sum(fvec_t *s) 
{
	unsigned int i,j;
	float tmp = 0.0f;
	for (i=0; i < s->channels; i++)
		for (j=0; j < s->length; j++)
			tmp += s->data[i][j];
	return tmp;
}


float vec_max(fvec_t *s) 
{
	unsigned int i,j;
	float tmp = 0.0f;
	for (i=0; i < s->channels; i++)
		for (j=0; j < s->length; j++)
			tmp = (tmp > s->data[i][j])? tmp : s->data[i][j];
	return tmp;
}

float vec_min(fvec_t *s) 
{
	unsigned int i,j;
	float tmp = 0.0f;
	for (i=0; i < s->channels; i++)
		for (j=0; j < s->length; j++)
			tmp = (tmp < s->data[i][j])? tmp : s->data[i][j]  ;
	return tmp;
}

void vec_shift(fvec_t *s)
{
	unsigned int i,j;
	float tmp = 0.0f;
	for (i=0; i < s->channels; i++)
		for (j=0; j < s->length / 2 ; j++) {
			tmp = s->data[i][j];
			s->data[i][j] = s->data[i][j+s->length/2];
			s->data[i][j+s->length/2] = tmp;
		}
}


/*
 * inspired from http://www.musicdsp.org/showone.php?id=27
 * Resonant IIR lowpass (12dB/oct) by Olli Niemitalo
 *
 * (was in double)
 */
/* resofreq becames resofreq/samplerate */
void aubio_lowpassresoiir(fvec_t * t, float resofreq, float amp) {
	float w = 2.0*PI*resofreq; // Pole angle
	float q = 1.0-w/(2.0*(amp+0.5/(1.0+w))+w-2.0); // Pole magnitude
	float r = q*q;
	float c = r+1.0-2.0*cos(w)*q;
	float vibrapos = 0.;
	float vibraspeed = 0.;
	float temp = 0.;
	unsigned int i,streamofs;

	/* Main loop */
	for (i=0;i<t->channels;i++) {
		for (streamofs = 0; streamofs < t->length; streamofs++) {

			/* Accelerate vibra by signal-vibra, multiplied by lowpasscutoff */
			vibraspeed += (t->data[i][streamofs] - vibrapos) * c;

			/* Add velocity to vibra's position */
			vibrapos += vibraspeed;

			/* Attenuate/amplify vibra's velocity by resonance */
			vibraspeed *= r;

			/* Check clipping */
			temp = vibrapos;
			//if (temp > 1.0) temp = 1.0; else if (temp < -1.0) temp = -1.0;

			/* Store new value */
			//tostream[streamofs] = temp;
			/* in place new value */
			t->data[i][streamofs] = temp;
		}
	}
}



#define ELEM_SWAP(a,b) { register float t=(a);(a)=(b);(b)=t; }

float quick_select(float arr[], int n) 
{
    int low, high ;
    int median;
    int middle, ll, hh;

    low = 0 ; high = n-1 ; median = (low + high) / 2;
    for (;;) {
        if (high <= low) /* One element only */
            return arr[median] ;

        if (high == low + 1) {  /* Two elements only */
            if (arr[low] > arr[high])
                ELEM_SWAP(arr[low], arr[high]) ;
            return arr[median] ;
        }

    /* Find median of low, middle and high items; swap into position low */
    middle = (low + high) / 2;
    if (arr[middle] > arr[high])    ELEM_SWAP(arr[middle], arr[high]) ;
    if (arr[low] > arr[high])       ELEM_SWAP(arr[low], arr[high]) ;
    if (arr[middle] > arr[low])     ELEM_SWAP(arr[middle], arr[low]) ;

    /* Swap low item (now in position middle) into position (low+1) */
    ELEM_SWAP(arr[middle], arr[low+1]) ;

    /* Nibble from each end towards middle, swapping items when stuck */
    ll = low + 1;
    hh = high;
    for (;;) {
        do ll++; while (arr[low] > arr[ll]) ;
        do hh--; while (arr[hh]  > arr[low]) ;

        if (hh < ll)
        break;

        ELEM_SWAP(arr[ll], arr[hh]) ;
    }

    /* Swap middle item (in position low) back into correct position */
    ELEM_SWAP(arr[low], arr[hh]) ;

    /* Re-set active partition */
    if (hh <= median)
        low = ll;
        if (hh >= median)
        high = hh - 1;
    }
}

#undef ELEM_SWAP
