/*
	 Copyright (C) 2003 Paul Brossier

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License as published by
	 the Free Software Foundation; either version 2 of the License, or
	 (at your option) any later version.

	 This program is distributed in the hope that it will be useful,
	 but WITHOUT ANY WARRANTY; without even the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	 GNU General Public License for more details.

	 You should have received a copy of the GNU General Public License
	 along with this program; if not, write to the Free Software
	 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/** @file
 *  various math functions
 */

#ifndef MATHUTILS_H
#define MATHUTILS_H

#include "fft.h"
#include "sample.h"

#define PI 				(M_PI)
#define TWO_PI 		(M_PI*2.)

/* aliases to math.h functions */
#define EXP				expf
#define COS				cosf
#define SIN				sinf
#define ABS				fabsf
#define POW				powf
#define SQRT			sqrtf
#define LOG10			log10f

/* aliases to complex.h functions */
#define EXPC			cexpf
#define CEXPC			cexp
#define ARGC			cargf
#define ABSC			cabsf
#define REAL			crealf
#define IMAG			cimagf

/* handy shortcuts */
#define DB2LIN(g) (POW(10.0f,(g)*0.05f))
#define LIN2DB(v) (20.0f*LOG10(v))
#define SQR(_a) 	(_a*_a)

/** Window types 
 * 
 * inspired from 
 *
 *  - dafx : http://profs.sci.univr.it/%7Edafx/Final-Papers/ps/Bernardini.ps.gz
 *  - freqtweak : http://freqtweak.sf.net/
 *  - extace : http://extace.sf.net/
 */
typedef enum {
	rectangle,
	hamming,
	hanning,
	hanningz,
	blackman,
	blackman_harris,
	gaussian,
	welch,
	parzen
} window_type_t;

typedef struct _spectral_peak_t spectral_peak_t;

struct _spectral_peak_t {
	int loc;
	float val;
};

/** create window */
void window(float *w, int size, window_type_t wintype);
/** get norm from spectrum */
void getnorm(float * norm, fft_data_t * spectrum, int size);
/** get phase from spectrum */
void getphase(float * phase, fft_data_t * spectrum, int size);

/**
 * Calculate interpolated frequencies
 *
 * calculate the interpolated values of location (bin),
 * phase and magnitude by cubic interpolation
 *
 *    from Xavier Serra, DAFx book
 *
 * - input data (from peak_pick)
 *   - r: modulus of the FFT
 *   - phi: phase of the FFT
 *   - ft.loc: peak locations (bin)
 *   - ft.val: peak magnitudes
 * - parameters
 *   - N: size of the FFT 
 * - output data
 *   - iftloc: interpolated location (bin) in float
 *   - iftval: interpolated magnitude
 *   - iftphase: interpolated phase
 *
 * NOT optimized at all
 *
 * Note: this is brute force implementation and would require 
 * some profiling
 */
void interpolfreq(
		const float * r, const float * phi, const int N, 
		const spectral_peak_t * ft,
		const int nbins,
		const int zp, 
		float * iftloc, float * iftphase, float * iftval
		);

/** principal argument
 *
 * mod(phase+PI,-TWO_PI)+PI 
 */
float unwrap2pi (float phase);

/** sort 'ft.bins' and ft.loc' arrays */
void sort_peak(spectral_peak_t * ft, int nbins);
void sort_array(fvec_t * t);

/** picking 'nbins' peaks above threshold 
 * in order of frequency */
void peak_pick(
		/** data */
		const float * r, 
		/** number of bins to peak */
		int nbins, 
		/** total number of bins */
		const int N, 
		/** low norm threshold (0.1)*/
		const int threshold,
		/** output data */
		spectral_peak_t * ft
		);

/** mean 
 * calculates the mean of a vector
 */
float vec_mean(fvec_t *s);
/** max
 * returns the max of a vector
 */
float vec_max(fvec_t *s);
/** min
 * returns the min of a vector
 */
float vec_min(fvec_t *s);
/** 
 * implement 'fftshift' like function
 * 
 * a[0]...,a[n/2],a[n/2+1],...a[n]
 * 
 * becomes
 * 
 * a[n/2+1],...a[n],a[0]...,a[n/2]
 */
void vec_shift(fvec_t *s);

float vec_sum(fvec_t *s);


float vec_median(fvec_t *t);


void aubio_lowpassresoiir(fvec_t * input, float resofreq, float amp);

float quick_select(float arr[], int n);
#endif
