/*
	 Copyright (C) 2003 Paul Brossier

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License as published by
	 the Free Software Foundation; either version 2 of the License, or
	 (at your option) any later version.

	 This program is distributed in the hope that it will be useful,
	 but WITHOUT ANY WARRANTY; without even the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	 GNU General Public License for more details.

	 You should have received a copy of the GNU General Public License
	 along with this program; if not, write to the Free Software
	 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	 
*/

#include "aubio_priv.h"
#include "fft.h"
#include "mathutils.h"
#include "phasevoc.h"

struct _aubio_pvoc_t {
	int win_s;
	int hop_s;
	int channels;
	aubio_fft_t * spectrum;
	fvec_t * synth;		 //current output grain [win_s]
	fvec_t * synthold; //last input frame 		[win_s-hop_s]
	fft_data_t ** spec;		 //current spectrum 		[win_s]
	fvec_t * data;		 //current input grain 	[win_s]
	fvec_t * dataold;  //last input frame 		[win_s-hop_s]
	float * w;
};


/** Memory allocation */
static aubio_pvoc_t * aubio_pvoc_malloc (int win_s, int hop_s, int channels);
/** object deletion */
static void aubio_pvoc_free (aubio_pvoc_t *pv);
/** returns data and dataold slided by hop_s */
static void aubio_pvoc_swapbuffers(
		float * data,
		float * dataold,
		const float * datanew,
		int win_s, int hop_s);
/** do additive synthesis from 'old' and 'cur' */
static void aubio_pvoc_addsynth(
		const float * synth,
		float * synthold,
		float * synthnew, 
		int win_s, int hop_s);


// process pv->datanew 
void aubio_pvoc_do(aubio_pvoc_t *pv, fvec_t * datanew, cvec_t *fftgrain) {
	unsigned int i,j;
	for (i=0; i<pv->channels; i++) {
		// slide 
		aubio_pvoc_swapbuffers(pv->data->data[i],
				pv->dataold->data[i],
				datanew->data[i],
				pv->win_s,pv->hop_s);
		// windowing
		for (j=0; j<pv->win_s; j++) 
			pv->data->data[i][j] *= pv->w[j];
		// fftshift
		vec_shift(pv->data);
		// calculate fft
		aubio_fft_do(pv->spectrum, 
				pv->data->data[i], 
				pv->spec[i], 
				pv->win_s);
		// get norm and phase
		getnorm(fftgrain->norm[i], pv->spec[i], pv->win_s);
		getphase(fftgrain->phas[i], pv->spec[i], pv->win_s);
	}
}

// write to synthnew
void aubio_pvoc_rdo(aubio_pvoc_t *pv,cvec_t * fftgrain, fvec_t * synthnew) {
	unsigned int i, j;
	for (i=0; i<pv->channels; i++) {
		for (j=0; j<pv->win_s/2; j++) {
			pv->spec[i][j] = fftgrain->norm[i][j]*cexp(I*unwrap2pi(fftgrain->phas[i][j]));
			pv->spec[i][pv->win_s-j] = 0.;
		}
		aubio_fft_rdo(pv->spectrum,
				pv->spec[i],
				pv->synth->data[i], 
				pv->win_s);
		vec_shift(pv->synth);
		aubio_pvoc_addsynth(pv->synth->data[i],
				pv->synthold->data[i],
				synthnew->data[i],
				pv->win_s,pv->hop_s);
	}
}

void del_aubio_pvoc(aubio_pvoc_t *pv) {
	aubio_pvoc_free(pv);
}

aubio_pvoc_t * new_aubio_pvoc (int win_s, int hop_s, int channels) {
	aubio_pvoc_t * pv = aubio_pvoc_malloc(win_s, hop_s, channels);
	window(pv->w,pv->win_s,hanningz);
	return pv;
}

static aubio_pvoc_t * aubio_pvoc_malloc (int win_s, int hop_s, int channels) {
	unsigned int i;
	aubio_pvoc_t * pv = AUBIO_NEW(aubio_pvoc_t);
	pv->spectrum = new_aubio_fft(win_s);

	// remember old
	pv->data = new_fvec (win_s, channels);
	pv->synth = new_fvec (win_s, channels);

	// new input output
	pv->dataold = new_fvec  (win_s-hop_s, channels);
	pv->synthold = new_fvec (win_s-hop_s, channels);
	pv->w = AUBIO_ARRAY(float,win_s);
	
	pv->spec = AUBIO_ARRAY(fft_data_t*,channels);
	for (i=0; i<channels; i++) 
		pv->spec[i] = AUBIO_ARRAY(fft_data_t,win_s);

	pv->channels = channels;
	pv->hop_s = hop_s;
	pv->win_s = win_s;
	return pv;
}

static void aubio_pvoc_free (aubio_pvoc_t *pv) {
	unsigned int i;
	del_aubio_fft(pv->spectrum);
	del_fvec(pv->dataold);
	del_fvec(pv->data);
	del_fvec(pv->synthold);
	del_fvec(pv->synth);
	for (i=0; i< pv->channels; i++) {
		AUBIO_FREE(pv->spec[i]);
	}
	AUBIO_FREE(pv->spec);
	AUBIO_FREE(pv->w);
	AUBIO_FREE(pv);
}

static void aubio_pvoc_swapbuffers(float * data, float * dataold, const float * datanew, 
		int win_s, int hop_s)
{
	unsigned int i;
  for (i=0;i<win_s-hop_s;i++)
		data[i] = dataold[i];
	for (i=0;i<hop_s;i++)
		data[win_s-hop_s+i] = datanew[i];
	for (i=0;i<win_s-hop_s;i++)
		dataold[i] = data[i+hop_s];
}

static void aubio_pvoc_addsynth(const float * synth, float * synthold, float * synthnew, 
		int win_s, int hop_s)
{
	unsigned int i;
 // add new synth to old one and put result in synthnew
 for (i=0;i<hop_s;i++) 						
	 synthnew[i] = synthold[i]+synth[i];
 // shift synthold
 for (i=0;i<win_s-2*hop_s;i++)	
	 synthold[i] = synthold[i+hop_s];
 // erase last frame in synthold
 for (i=win_s-hop_s;i<win_s;i++)	
	 synthold[i-hop_s]=0.;
 // additive synth
 for (i=0;i<win_s-hop_s;i++)	
	 synthold[i] += synth[i+hop_s];
}

